open State
open Syntax
open Log
open Flags
open Eproverstate

let eeqoequiv = ref false
let eproverperiod = ref 10
let eprovertimeout = ref 1
let foffiles = ref false


let plit_fo_str a =
  if a > 0 then
    "p" ^ (string_of_int a)
  else
    "~ p" ^ (string_of_int (- a))

let rec printf_fotrm s m =
  match m with
  | FOVar(x) -> Printf.fprintf s "%s" x
  | FOFun(f,[]) -> Printf.fprintf s "c%s" f
  | FOFun(f,(a::al)) ->
      begin
	Printf.fprintf s "f%s(" f;
	printf_fotrm s a;
	List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
	Printf.fprintf s ")";
      end


let rec printf_fof s m =
  match m with
  | FOEq(t1,t2) ->
      Printf.fprintf s "(";
      printf_fotrm s t1;
      Printf.fprintf s " = ";
      printf_fotrm s t2;
      Printf.fprintf s ")";
  | FOPred(p,[]) ->
      Printf.fprintf s "q%s" p;
  | FOPred(p,(a::al)) ->
      Printf.fprintf s "r%s(" p;
      printf_fotrm s a;
      List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
      Printf.fprintf s ")";
  | FOImp(m1,m2) ->
      Printf.fprintf s "(";
      printf_fof s m1;
      Printf.fprintf s " => ";
      printf_fof s m2;
      Printf.fprintf s ")";
  | FOEqu(m1,m2) ->
      Printf.fprintf s "(";
      printf_fof s m1;
      Printf.fprintf s " <=> ";
      printf_fof s m2;
      Printf.fprintf s ")";
  | FOAll(x,m1) ->
      Printf.fprintf s "( ! [%s] : " x;
      printf_fof s m1;
      Printf.fprintf s ")";
  | FOFal ->
      Printf.fprintf s "$false"


let printf_fof_file s b =
  let axc = ref 0 in
	   (*** Send all propositional clauses we've sent to MiniSat. This contains the propositional structure of the problem. ***)
  List.iter (fun c ->
    match c with
    | (l1::ll) ->
	incr axc;
	Printf.fprintf s "fof(ax%d,axiom,( %s " (!axc) (plit_fo_str l1);
	List.iter (fun l2 -> Printf.fprintf s " | %s" (plit_fo_str l2)) ll;
	Printf.fprintf s ")).\n";
    | [] -> () (*** should never happen ***)
	  )
    !clauses;
	   (*** Relate Some propositional literals to first order formulas. ***)
	   (*** First the ones that are really propositional. ***)
  List.iter (fun (p,fof) ->
    incr axc;
    Printf.fprintf s "fof(ax%d,axiom,( %s <=> " (!axc) (plit_fo_str p);
    printf_fof s fof;
    Printf.fprintf s ")).\n"
      )
    (!prop_fo_forms);
	   (*** Finally the ones that are specifically viewed as first order via the current type b. ***)
  List.iter (fun (p,fof) ->
    incr axc;
    Printf.fprintf s "fof(ax%d,axiom,( %s <=> " (!axc) (plit_fo_str p);
    printf_fof s fof;
    Printf.fprintf s ")).\n"
      )
    (Hashtbl.find_all stp_fo_forms b)


(*** Output FOF as Sexpr -- just for debugging BEGIN ***)
let rec printf_fstrm s m =
  match m with
  | FOVar(x) -> Printf.fprintf s "|%s|" x
  | FOFun(f,[]) -> Printf.fprintf s "|%s|" f
  | FOFun(f,(a::al)) ->
      begin
	Printf.fprintf s "(|%s| " f;
	printf_fstrm s a;
	List.iter (fun t -> Printf.fprintf s " "; printf_fstrm s t) al;
	Printf.fprintf s ")";
      end


let rec printf_fs s m =
  match m with
  | FOEq(t1,t2) ->
      Printf.fprintf s "(= ";
      printf_fstrm s t1;
      Printf.fprintf s " ";
      printf_fstrm s t2;
      Printf.fprintf s ")";
  | FOPred(p,[]) ->
      Printf.fprintf s "|%s|" p;
  | FOPred(p,(a::al)) ->
      Printf.fprintf s "(|%s| " p;
      printf_fstrm s a;
      List.iter (fun t -> Printf.fprintf s " "; printf_fstrm s t) al;
      Printf.fprintf s ")";
  | FOImp(m1,m2) ->
      Printf.fprintf s "(=> ";
      printf_fs s m1;
      Printf.fprintf s " ";
      printf_fs s m2;
      Printf.fprintf s ")";
  | FOEqu(m1,m2) ->
      Printf.fprintf s "(<=> ";
      printf_fs s m1;
      Printf.fprintf s " ";
      printf_fs s m2;
      Printf.fprintf s ")";
  | FOAll(x,m1) ->
      Printf.fprintf s "(ALL |%s| " x;
      printf_fs s m1;
      Printf.fprintf s ")";
  | FOFal ->
      Printf.fprintf s "FALSE"


let printf_fs_file s b =
  let axc = ref 0 in
	   (*** Send all propositional clauses we've sent to MiniSat. This contains the propositional structure of the problem. ***)
  List.iter (fun c ->
    match c with
    | (l1::ll) ->
	incr axc;
	Printf.fprintf s "(AXIOM %d (|%s|" (!axc) (plit_fo_str l1);
	List.iter (fun l2 -> Printf.fprintf s " |%s|" (plit_fo_str l2)) ll;
	Printf.fprintf s "))\n";
    | [] -> () (*** should never happen ***)
	  )
    !clauses;
	   (*** Relate Some propositional literals to first order formulas. ***)
	   (*** First the ones that are really propositional. ***)
  List.iter (fun (p,fof) ->
    incr axc;
    Printf.fprintf s "(EAXIOM %d |%s| " (!axc) (plit_fo_str p);
    printf_fs s fof;
    Printf.fprintf s ")\n"
      )
    (!prop_fo_forms);
	   (*** Finally the ones that are specifically viewed as first order via the current type b. ***)
  List.iter (fun (p,fof) ->
    incr axc;
    Printf.fprintf s "(EAXIOM %d |%s| " (!axc) (plit_fo_str p);
    printf_fs s fof;
    Printf.fprintf s ")\n"
      )
    (Hashtbl.find_all stp_fo_forms b)
(*** Output FOF as Sexpr -- just for debugging END ***)



let call_e b =
	   if (!verbosity > 4) then (Printf.printf "Calling E on FOF for type %s\n" (stp_str b); flush stdout);
	   if (!foffiles) then
	     begin
	       let foffilename = "/tmp/fof" ^ (string_of_float (Unix.time ())) ^ "p" in
	       let foffile = open_out foffilename in
	       printf_fof_file foffile b;
	       close_out foffile;
	       Printf.printf "./E/PROVER/eprover -s --auto --cpu-limit=1 --tptp3-in < %s\n" foffilename;
	       Unix.sleep 1; (*** sleep for a second so the names will be unique ***)
	     end;
(***
 ***)
	   let etimeoutnow =
             match (!timeout) with
             | Some _ -> min (int_of_float (remaining_time ())) (!eprovertimeout)
             | None -> !eprovertimeout
           in
	   if (etimeoutnow > 0) then
	     begin
	       if (!verbosity > 4) then Printf.printf "Calling E for %d seconds.\n" etimeoutnow;
	       let (myout,myin,myerr) = Unix.open_process_full ((!Config.eprover) ^ " -s --auto --cpu-limit=" ^ (string_of_int etimeoutnow) ^ "  --tptp3-in") [| |] in
               printf_fof_file myin b;
	       close_out myin;
	       if (!verbosity > 4) then (Printf.printf "About to read result from eprover\n"; flush stdout);
	       try
		 while true do
		   match (input_line myout) with
		   | "# SZS status Unsatisfiable" ->
                       if (!foffiles) then
			 begin
			   let foffilename = "/tmp/fof" ^ (string_of_float (Unix.time ())) ^ "p" in
			   let foffile = open_out foffilename in
			   printf_fof_file foffile b;
			   close_out foffile;
			   let fsfilename = "/tmp/fs" ^ (string_of_float (Unix.time ())) ^ ".lisp" in
			   let fsfile = open_out fsfilename in
			   printf_fs_file fsfile b;
			   close_out fsfile;
			   Printf.printf "./E/PROVER/eprover -s --auto --cpu-limit=1 --tptp3-in < %s\n" foffilename;
			 end;
                       ignore(Unix.close_process_full(myout,myin,myerr));
		       raise (Unsatisfiable None)
		   | _ -> ()
		 done
	       with End_of_file -> ignore(Unix.close_process_full(myout,myin,myerr))
	     end

(*** Remember if a is FO, so I can send this to E. Send it to E every E_PERIOD times. ***)
let new_possible_fo_formula m a =
 try
   match trm_to_foform_stp m (!eeqoequiv) with
   | (fof,Some b) ->
       if (not (Hashtbl.mem fo_types_h b)) then
	 begin
	   if (!verbosity > 4) then (Printf.printf "New FO type: %s\n" (stp_str b); flush stdout);
	   Hashtbl.add fo_types_h b ();
	   fo_types := b::!fo_types;
	   Hashtbl.add fo_ecounter b (ref 0);
	 end;
       if (!verbosity > 4) then (Printf.printf "Adding FOF for type %s: %s\n" (stp_str b) (trm_str m); flush stdout);
       Hashtbl.add stp_fo_forms b (a,fof);
       let r = Hashtbl.find fo_ecounter b in
       incr r;
       if (!r >= !eproverperiod) then (r := 0; call_e b)
   | (fof,None) ->
       if (!verbosity > 4) then (Printf.printf "Adding prop FOF: %s\n" (trm_str m); flush stdout);
       prop_fo_forms := ((a,fof)::!prop_fo_forms)
 with NotFO -> ()


let setup_eprover () =
  if get_bool_flag "USE_E" then
    if !Config.eprover = "" then print_endline "Warning: disabling eprover!"
    else Atom.new_atom_action := new_possible_fo_formula;

  eeqoequiv := get_bool_flag "E_EQO_EQUIV";
  eproverperiod := get_int_flag "E_PERIOD";
  eprovertimeout := get_int_flag "E_TIMEOUT"
