(* File: refut.ml *)
(* Author: Chad E Brown *)
(* Created: October 2010 *)

open Syntax

type clause = int list

type ruleinfo =
 (*** NegPropRule(m) For rules requiring one negative formula in the head, except negated forall ***)
    NegPropRule of trm
 (*** PosPropRule(m) For rules requiring one positive formula in the head, except forall ***)
  | PosPropRule of trm
 (*** InstRule(a,m,n) For instantiating a Ap(Forall(a),m) with n:a ***)
  | InstRule of stp * trm * trm
 (*** FreshRule(a,m,x) For producing a witness x for Ap(Neg,Ap(Forall(a),m)) ***)
  | FreshRule of stp * trm * string
 (*** MatingRule(plit,nlit) For mating rule, only save the literals here ***)
  | MatingRule of int * int
 (*** ConfrontationRule(plit,nlit) For confrontation rule, only save the literals here ***)
  | ConfrontationRule of int * int
 (*** ChoiceRule(eps,pred) ***)
  | ChoiceRule of trm * trm
 (*** Known(lit,coqname,stp list) ***)
  | Known of int * string * stp list

type refut =
    NegImpR of trm * trm * refut
  | ImpR of trm * trm * refut * refut
  | FalseR
  | NegReflR
  | NegAllR of stp * trm * string * refut
  | EqFR of stp * stp * trm * trm * refut
  | NegEqFR of stp * stp * trm * trm * refut
  | EqOR of trm * trm * refut * refut
  | NegEqOR of trm * trm * refut * refut
  | SearchR of clause list * (clause -> ruleinfo)
  | AssumptionConflictR

(*** For printing a ruleinfo - helpful for debugging. - Chad, Oct 2010 ***)
let ruleinfo_str s =
  match s with
  | NegPropRule(m) -> ("NegPropRule(" ^ (trm_str m) ^ ")")
  | PosPropRule(m) -> ("PosPropRule(" ^ (trm_str m) ^ ")")
  | InstRule(a,m,n) -> ("InstRule(" ^ (stp_str a) ^ "," ^ (trm_str m) ^ "," ^ (trm_str n) ^ ")")
  | FreshRule(a,m,x) -> ("FreshRule(" ^ (stp_str a) ^ "," ^ (trm_str m) ^ "," ^ x ^ ")")
  | MatingRule(plit,nlit) -> "MatingRule(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | ConfrontationRule(plit,nlit) -> "ConfrontationRule(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | ChoiceRule(eps,pred) -> "ChoiceRule(" ^ (trm_str eps) ^ "," ^ (trm_str pred) ^ ")"
  | Known(l,cname,al) -> "Known(" ^ (string_of_int l) ^ "," ^ cname ^ ")"

(*** For printing the refutation - helpful for debugging. - Chad, Oct 2010 ***)
let rec print_refut r =
  match r with
  | NegImpR(m,n,r1) ->
      Printf.printf "NegImpR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1
  | ImpR(m,n,r1,r2) ->
      Printf.printf "ImpR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | FalseR ->
      Printf.printf "FalseR\n"
  | NegReflR ->
      Printf.printf "NegReflR\n"
  | NegAllR(a,m,x,r) ->
      Printf.printf "NegAllR\n  fresh var:%s\n  stp: %s\n   body: %s\n" x (stp_str a) (trm_str m);
      print_refut r
  | EqFR(a,b,m,n,r) ->
      Printf.printf "EqFR\n  stp: (%s)   ->   (%s)\nleft: %s\n  right: %s\n" (stp_str a) (stp_str b) (trm_str m) (trm_str n);
  | NegEqFR(a,b,m,n,r) ->
      Printf.printf "NegEqFR\n  stp: (%s)   ->   (%s)\nleft: %s\n  right: %s\n" (stp_str a) (stp_str b) (trm_str m) (trm_str n);
  | EqOR(m,n,r1,r2) ->
      Printf.printf "EqOR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | NegEqOR(m,n,r1,r2) ->
      Printf.printf "NegEqOR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | AssumptionConflictR ->
      Printf.printf "AssumptionConflictR\n";
  | SearchR(cl,cr) ->
      Printf.printf "SearchR\n  clauses with search options:\n";
      List.iter
	(fun c ->
	  Printf.printf "%s\n" (String.concat " " (List.map string_of_int c));
	  List.iter
	    (fun z -> (Printf.printf "Atom %d: %s\n" (abs z) (trm_str (Atom.atom_to_trm (abs z)))))
	    c;
	  try Printf.printf "associated rule: %s\n" (ruleinfo_str (cr c))
	  with Not_found -> Printf.printf "no associated rule\n"
	)
	cl


(** Extraction of constants and types from different kinds of types. **)

(*List the constants that occur in a rule-info, as name-type pairs.
  NOTE the resulting list doesn't contain duplicate constants.*)
let rec consts_of_ruleinfo acc : ruleinfo -> ctx = function
    NegPropRule t
  | PosPropRule t
  | FreshRule (_, t, _) -> consts_of_trm acc t
  | InstRule (_, t1, t2)
  | ChoiceRule (t1, t2) ->
    let acc' = consts_of_trm acc t1
    in consts_of_trm acc' t2
  | ConfrontationRule (_, _)
  | MatingRule (_, _)
  | Known (_, _, _) -> acc

let fold_clause f =
  List.fold_right (fun z acc -> f acc (Atom.atom_to_trm (abs z)))

let fold_searchr ft fr cl cr =
  List.fold_right (fun c acc ->
    let acc' = fold_clause ft c acc in
    try fr acc' (cr c) with Not_found -> acc') cl

(*List the constants that occur in a refutation, as name-type pairs.
  NOTE the resulting list doesn't contain duplicate constants.*)
let rec consts_of_refut acc : refut -> ctx = function
    NegImpR (t1, t2, r)
  | EqFR (_, _, t1, t2, r)
  | NegEqFR (_, _, t1, t2, r) ->
    let acc' = consts_of_trm acc t1 in
    let acc'' = consts_of_trm acc' t2 in
    consts_of_refut acc'' r
  | NegAllR (stp, t, name, r) ->
    let acc' =
      begin
        print_endline name;
      try
        let stp' = List.assoc name acc
        in
          if stp' <> stp then failwith "Different types for same constant?"
          else acc
      with Not_found ->
        (name, stp) :: acc
      end in
    let acc'' = consts_of_trm acc' t in
    consts_of_refut acc'' r
  | EqOR (t1, t2, r1, r2)
  | NegEqOR (t1, t2, r1, r2)
  | ImpR (t1, t2, r1, r2) ->
    let acc' = consts_of_trm acc t1 in
    let acc'' = consts_of_trm acc' t2 in
    let acc''' = consts_of_refut acc'' r1 in
    consts_of_refut acc''' r2
  | FalseR
  | NegReflR
  | AssumptionConflictR -> acc
  | SearchR (cl, cr) -> fold_searchr consts_of_trm consts_of_ruleinfo cl cr acc

(*List the base types that are referenced in a rule-info.
  NOTE the resulting list doesn't contain duplicate elements.*)
let rec base_types_of_ruleinfo acc : ruleinfo -> string list = function
    NegPropRule t
  | PosPropRule t -> base_types_of_trm acc t
  | FreshRule (stp, t, _) ->
    let acc' = base_types acc stp
    in base_types_of_trm acc' t
  | InstRule (stp, t1, t2) ->
    let acc' = base_types_of_trm acc t1 in
    let acc'' = base_types_of_trm acc' t2
    in base_types acc'' stp
  | Known (_, _, stps) ->
    List.fold_right (fun x y -> base_types y x) stps acc
  | ChoiceRule (t1, t2) ->
    let acc' = base_types_of_trm acc t1
    in base_types_of_trm acc' t2
  | ConfrontationRule (_, _)
  | MatingRule (_, _) -> acc

(*List the base types that are referenced in a refutation.
  NOTE the resulting list doesn't contain duplicate elements.*)
let rec base_types_of_refut acc : refut -> string list = function
    NegImpR (t1, t2, r) ->
    let acc' = base_types_of_trm acc t1 in
    let acc'' = base_types_of_trm acc' t2 in
    base_types_of_refut acc'' r
  | EqFR (stp1, stp2, t1, t2, r)
  | NegEqFR (stp1, stp2, t1, t2, r) ->
    let acc' = base_types_of_trm acc t1 in
    let acc'' = base_types_of_trm acc' t2 in
    let acc3 = base_types acc'' stp1 in
    let acc4 = base_types acc3 stp2 in
    base_types_of_refut acc4 r
  | NegAllR (stp, t, _, r) ->
    let acc' = base_types acc stp in
    let acc'' = base_types_of_trm acc' t in
    base_types_of_refut acc'' r
  | EqOR (t1, t2, r1, r2)
  | NegEqOR (t1, t2, r1, r2)
  | ImpR (t1, t2, r1, r2) ->
    let acc' = base_types_of_trm acc t1 in
    let acc'' = base_types_of_trm acc' t2 in
    let acc''' = base_types_of_refut acc'' r1 in
    base_types_of_refut acc''' r2
  | FalseR
  | NegReflR
  | AssumptionConflictR -> acc
  | SearchR (cl, cr) -> fold_searchr base_types_of_trm base_types_of_ruleinfo cl cr acc

